import ctypes
from ctypes import windll, c_int, c_char, byref, c_long, \
    pointer, c_float, c_char_p, cdll

import sys
import platform

from pathlib import Path

try:
    # Check operating system and load library
    if platform.system() == "Linux":
        dllpath = Path("/usr/local/lib/libandor.so")
        _dll = cdll.LoadLibrary(str(dllpath))
    elif platform.system() == "Windows":
        dllpath = Path('C:\\Program Files\\Andor SDK')
        if platform.machine() == "AMD64":
            dllpath = dllpath.joinpath("Shamrock64/ShamrockCIF.dll")
        else:
            dllpath = dllpath.joinpath("Shamrock/ShamrockCIF.dll")
        _dll = windll.LoadLibrary(str(dllpath))

except Exception as e:
    raise Exception(f'Could not import Shamrock library: {str(e)}')


class ShamrockSDK:
    """
    Andor class which is meant to provide the Python version of the same
    functions that are defined in the Andor's SDK.
    """
    def __init__(self):
        '''
        'C:\\Program Files\\Andor SDK'
        
        Loads and initializes the hardware driver.
        Initializes local parameters

        Input:
            dllpath (string)   : The path where to find the sdk2 dlls
            control type (string) : either "camera", "shamrock" or "both". to be set to initialize all or part of the libraries
        '''

        print("Initializing Shamrock...", )
        error = _dll.ShamrockInitialize()

        nodevices = c_int()
        error = _dll.ShamrockGetNumberDevices(byref(nodevices))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        print(f'(Nr of devices: {nodevices.value})')
        
        # Initiate parameters - Shamrock
        self.NrPixels = 0

    def __del__(self):
        error = _dll.ShamrockClose()
        if error != 20202:
            raise IOError(ERROR_CODE[error])

    def close(self):
        error = _dll.ShamrockClose()
        if error != 20202:
            raise IOError(ERROR_CODE[error])

    def GetNumberDevicesSR(self):
        nodevices = c_int()
        error = _dll.ShamrockGetNumberDevices(byref(nodevices))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], nodevices.value

    def GetFunctionReturnDescriptionSR(self, error, MaxDescStrLen):
        error = c_int(error)
        MaxDescStrLen = c_int(MaxDescStrLen)
        description = c_char_p()
        error = _dll.ShamrockGetFunctionReturnDescription(error, description, MaxDescStrLen)
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], description.value

    #---- sdkeeprom functions

    def GetSerialNumberSR(self, device):
        device = c_int(device)
        serial = ctypes.create_string_buffer(128)
        error = _dll.ShamrockGetSerialNumber(device, byref(serial))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], serial.value

    def EepromGetOpticalParamsSR(self, device):
        device = c_int(device)
        FocalLength = c_float()
        AngularDeviation = c_float()
        FocalTilt = c_float()
        error = _dll.ShamrockEepromGetOpticalParams(device, byref(FocalLength),
                                                    byref(AngularDeviation), byref(FocalTilt))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], FocalLength.value, AngularDeviation.value, FocalTilt.value

    #---- sdkgrating functions
    def SetGratingSR(self, device, grating):
        device = c_int(device)
        grating = c_int(grating)
        error = _dll.ShamrockSetGrating(device, grating)
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error]

    def GetGratingSR(self, device):
        device = c_int(device)
        grating = c_int()
        error = _dll.ShamrockGetGrating(device, byref(grating))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], grating.value

    def WavelengthResetSR(self, device):
        device = c_int(device)
        error = _dll.ShamrockWavelengthReset(device)
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error]

    def GetNumberGratingsSR(self, device):
        device = c_int(device)
        noGratings = c_int()
        error = _dll.ShamrockGetNumberGratings(device, byref(noGratings))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], noGratings.value

    def GetGratingInfoSR(self, device, grating):
        device = c_int(device)
        grating = c_int(grating)
        Lines = c_float()
        Blaze = ctypes.create_string_buffer(128)
        Home = c_int()
        Offset = c_int()
        error = _dll.ShamrockGetGratingInfo(device, grating, 
                    byref(Lines), byref(Blaze), byref(Home), byref(Offset))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], Lines.value, Blaze.value.decode(), Home.value, Offset.value
    
    #---- sdkwavelength functions
    def SetWavelengthSR(self, device, wavelength):
        device = c_int(device)
        wavelength = c_float(wavelength)
        error = _dll.ShamrockSetWavelength(device, wavelength)
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error]
    
    def GetWavelengthSR(self, device):
        device = c_int(device)
        wavelength = c_float()
        error = _dll.ShamrockGetWavelength(device, byref(wavelength))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], wavelength.value
    
    def GetWavelengthLimitsSR(self, device, grating):
        device = c_int(device)
        grating = c_int(grating)
        minLambda = c_float()
        maxLambda = c_float()
        error = _dll.ShamrockGetWavelengthLimits(device, grating, byref(minLambda), byref(maxLambda))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], minLambda.value, maxLambda.value
    
    def GotoZeroOrderSR(self, device):
        device = c_int(device)
        error = _dll.ShamrockGotoZeroOrder(device)
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error]
    
    def AtZeroOrderSR(self, device):
        device = c_int(device)
        atZeroOrder = c_int()
        error = _dll.ShamrockAtZeroOrder(device, byref(atZeroOrder))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], atZeroOrder.value
    
    #---- sdkcalibration functions
    def SetNumberPixelsSR(self, device, NumberPixels):
        device = c_int(device)
        NumberPixels = c_int(NumberPixels)
        error = _dll.ShamrockSetNumberPixels(device, NumberPixels)
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        self.NrPixels = NumberPixels.value
        return ERROR_CODE[error]
    
    def SetPixelWidthSR(self, device, Width):
        device = c_int(device)
        Width = c_float(Width)
        error = _dll.ShamrockSetPixelWidth(device, Width)
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error]
    
    def GetNumberPixelsSR(self, device):
        device = c_int(device)
        NumberPixels = c_int()
        error = _dll.ShamrockGetNumberPixels(device, byref(NumberPixels))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        self.NrPixels = NumberPixels.value
        return ERROR_CODE[error], NumberPixels.value
    
    def GetPixelWidthSR(self, device):
        device = c_int(device)
        Width = c_float()
        error = _dll.ShamrockGetPixelWidth(device, byref(Width))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], Width.value
    
    def GetCalibrationSR(self, device, Npxls):
        device = c_int(device)
        CalibrationValues = (c_float*int(Npxls))()
        error = _dll.ShamrockGetCalibration(device, byref(CalibrationValues), int(Npxls))
        if error != 20202:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], CalibrationValues[:]


# ####################################################

# List of error codes
ERROR_CODE = {

    20201: "SHAMROCK_COMMUNICATION_ERROR",
    20202: "SHAMROCK_SUCCESS",
    20266: "SHAMROCK_P1INVALID",
    20267: "SHAMROCK_P2INVALID",
    20268: "SHAMROCK_P3INVALID",
    20269: "SHAMROCK_P4INVALID",
    20270: "SHAMROCK_P5INVALID",
    20275: "SHAMROCK_NOT_INITIALIZED",
    20292: "SHAMROCK_NOT_AVAILABLE"
}





import numpy as np
from enum import IntEnum
import ctypes
from ctypes.util import find_library
import platform
from qtpy import QtWidgets, QtCore
from easydict import EasyDict as edict

from pymodaq_utils.utils import ThreadCommand, find_dict_in_list_from_key_val
from pymodaq_gui.parameter.utils import iter_children

from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base, comon_parameters, main
from pymodaq.utils.data import DataFromPlugins, Axis, DataToExport

from pymodaq_plugins_andor.hardware.andor_sdk2 import sdk2

libpath = sdk2.dllpath
camera_list = sdk2.AndorSDK.GetCamerasInfo()


class Andor_Camera_ReadOut(IntEnum):
    """
        Enum class of readout modes.

        =============== =======================
        **Attributes**    **Type**
        *names*          string list of members
        =============== =======================
    """

    FullVertBinning = 0
    SingleTrack = 3
    MultiTrack = 1
    RandomTrack = 2
    Image = 4
    Cropped = 5


    @classmethod
    def names(cls):
        return [name for name, member in cls.__members__.items()]

class Andor_Camera_AcqMode(IntEnum):
    """
        Enum class of AcqModes modes.

        =============== =======================
        **Attributes**    **Type**
        *names*          string list of members
        =============== =======================
    """
    Single_Scan = 1
    Accumulate = 2
    Kinetics = 3
    Fast_Kinetics = 4
    Run_till_abort = 5


    @classmethod
    def names(cls):
        return [name for name, member in cls.__members__.items()]


class DAQ_2DViewer_AndorCCD(DAQ_Viewer_base):
    """
        Base class for Andor CCD camera and Shamrock spectrometer


        =============== ==================
        **Attributes**   **Type**

        =============== ==================

        See Also
        --------
        utility_classes.DAQ_Viewer_base
    """
    callback_signal = QtCore.Signal(int)
    hardware_averaging = True #will use the accumulate acquisition mode if averaging is neccessary
    params = comon_parameters+[
        {'title': 'Dll library:', 'name': 'andor_lib', 'type': 'browsepath', 'value': str(libpath)},
        
        {'title': 'Camera Settings:', 'name': 'camera_settings', 'type': 'group', 'expanded': True, 'children': [
            {'title': 'Camera Models:', 'name': 'camera_model', 'type': 'list',
                'limits': [f"{cam['model']}-{cam['serial']}" for cam in camera_list]},

            {'title': 'Readout Modes:', 'name': 'readout', 'type': 'list', 'limits': Andor_Camera_ReadOut.names()[0:-1],
                                            'value': 'FullVertBinning'},
            {'title': 'Readout Settings:', 'name': 'readout_settings', 'type': 'group', 'children':[

                {'title': 'single Track Settings:', 'name': 'st_settings', 'type': 'group', 'visible': False, 'children':[
                    {'title': 'Center pixel:', 'name': 'st_center', 'type': 'int', 'value': 1 , 'default':1, 'min':1},
                    {'title': 'Height:', 'name': 'st_height', 'type': 'int', 'value': 1 , 'default':1, 'min':1},
                ]},    
                {'title': 'Multi Track Settings:', 'name': 'mt_settings', 'type': 'group', 'visible': False, 'children':[
                    {'title': 'Ntrack:', 'name': 'mt_N', 'type': 'int', 'value': 1 , 'default':1, 'min':1},
                    {'title': 'Height:', 'name': 'mt_height', 'type': 'int', 'value': 1 , 'default':1, 'min':1},
                    {'title': 'Offset:', 'name': 'mt_offset', 'type': 'int', 'value': 1 , 'default':1, 'min':0},
                    {'title': 'Bottom:', 'name': 'mt_bottom', 'type': 'int', 'value': 1 , 'default':1, 'min':0, 'readonly': True},
                    {'title': 'Gap:', 'name': 'mt_gap', 'type': 'int', 'value': 1 , 'default':1, 'min':0, 'readonly': True},
                ]},
                {'title': 'Image Settings:', 'name': 'image_settings', 'type': 'group', 'visible': False, 'children':[
                    {'title': 'Binning along x:', 'name': 'bin_x', 'type': 'int', 'value': 1, 'default': 1, 'min': 1},
                    {'title': 'Binning along y:', 'name': 'bin_y', 'type': 'int', 'value': 1, 'default': 1, 'min': 1},
                    {'title': 'Start x:', 'name': 'im_startx', 'type': 'int', 'value': 1 , 'default':1, 'min':0},
                    {'title': 'End x:', 'name': 'im_endx', 'type': 'int', 'value': 1024 , 'default':1024, 'min':0},
                    {'title': 'Start y:', 'name': 'im_starty', 'type': 'int', 'value': 1 , 'default':1, 'min':1},
                    {'title': 'End y:', 'name': 'im_endy', 'type': 'int', 'value': 256, 'default':256, 'min':1,},
                    ]},   
            ]},            
            {'title': 'Exposure (ms):', 'name': 'exposure', 'type': 'float', 'value': 0.01 , 'default':0.01, 'min': 0},
            
            {'title': 'Image size:', 'name': 'image_size', 'type': 'group', 'children':[
                {'title': 'Nx:', 'name': 'Nx', 'type': 'int', 'value': 0, 'default':0 , 'readonly': True},
                {'title': 'Ny:', 'name': 'Ny', 'type': 'int', 'value': 0 , 'default':0 , 'readonly': True},
                ]},
            
            {'title': 'Shutter Settings:', 'name': 'shutter', 'type': 'group', 'children':[
                {'title': 'Open Shutter on:', 'name': 'shutter_type', 'type': 'list', 'value': 'high', 'limits': ['low', 'high']},
                {'title': 'Shutter mode:', 'name': 'shutter_mode', 'type': 'list', 'value': 'Auto', 'limits': ['Auto', 'Always Opened', 'Always Closed', ]},
                {'title': 'Closing time (ms):', 'name': 'shutter_closing_time', 'type': 'int', 'value': 0, 'tip': 'millisecs it takes to close'},
                {'title': 'Opening time (ms):', 'name': 'shutter_opening_time', 'type': 'int', 'value': 10, 'tip': 'millisecs it takes to open'},
                ]},
            {'title': 'Temperature Settings:', 'name': 'temperature_settings', 'type': 'group', 'children': [
                {'title': 'Set Point:', 'name': 'set_point', 'type': 'float', 'value': -60, 'default': -60},
                {'title': 'Current value:', 'name': 'current_value', 'type': 'float', 'value': 0, 'default': 0,
                 'readonly': True},
                {'title': 'Locked:', 'name': 'locked', 'type': 'led', 'value': False, 'default': False,
                 'readonly': True},
            ]},
        ]},
        ]

    def ini_attributes(self):

        self.camera_controller: sdk2.AndorSDK() = None

        self.x_axis = None
        self.y_axis = None
        self.camera_controller = None
        self.data = None
        self.CCDSIZEX, self.CCDSIZEY = (None, None)
        self.data_pointer = None
        self.camera_done = False
        self.acquirred_image = None
        self.callback_thread = None
        self.Naverage = None
        self.data_shape = None  # 'Data2D' if sizey != 1 else 'Data1D'
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.updated_timer)

    def commit_settings(self, param):
        """
            | Activate parameters changes on the hardware from parameter's name.
            |

            =============== ================================    =========================
            **Parameters**   **Type**                           **Description**
            *param*          instance of pyqtgraph parameter    The parameter to activate
            =============== ================================    =========================

            Three profile of parameter :
                * **bin_x** : set binning camera from bin_x parameter's value
                * **bin_y** : set binning camera from bin_y parameter's value
                * **set_point** : Set the camera's temperature from parameter's value.

        """
        try:
            if param.name() == 'set_point':
                self.camera_controller.SetTemperature(param.value())

            elif param.name() == 'readout' or param.name() in iter_children(self.settings.child('camera_settings', 'readout_settings')):
                self.update_read_mode()
                
            elif param.name() == 'exposure':
                self.camera_controller.SetExposureTime(param.value() / 1000) #temp should be in s
                (err, timings) = self.camera_controller.GetAcquisitionTimings()
                self.settings.child('camera_settings', 'exposure').setValue(timings['exposure']*1000)
                QtWidgets.QApplication.processEvents()
                #self.get_exposure_ms()

            elif param.name() in iter_children(self.settings.child('camera_settings', 'shutter'), []):
                self.set_shutter()

            elif param.name() in iter_children(self.settings.child('camera_settings', 'readout_settings', 'image_settings')):
                if self.settings.child('camera_settings', 'readout').value() == 'Image':
                    self.set_image_area()
            pass
        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [str(e), 'log']))

    def emit_data(self):
        """
            Fonction used to emit data obtained by callback.

            See Also
            --------
            daq_utils.ThreadCommand
        """
        try:
            self.ind_grabbed += 1

            sizey = self.settings.child('camera_settings', 'image_size', 'Ny').value()
            sizex = self.settings.child('camera_settings', 'image_size', 'Nx').value()
            self.camera_controller.GetAcquiredDataNumpy(self.data_pointer, sizex * sizey)
            self.dte_signal.emit(DataToExport('Camera',
                                              data=[DataFromPlugins(name='Camera',
                                                                    data=[np.squeeze(
                                                                        self.data.reshape((sizey, sizex)).astype(float))],
                                                                    dim=self.data_shape)]))
            QtWidgets.QApplication.processEvents()  # here to be sure the timeevents are executed even if in continuous grab mode

        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [str(e), 'log']))


    def update_read_mode(self):
        read_mode = Andor_Camera_ReadOut[self.settings.child('camera_settings', 'readout').value()].value
        err = self.camera_controller.SetReadMode(read_mode)
        if err != 'DRV_SUCCESS':
            self.emit_status(ThreadCommand('Update_Status',[err,'log']))
        else:
            self.settings.child('camera_settings', 'readout_settings').show()
            if read_mode == 0:#FVB:
                self.settings.child('camera_settings', 'readout_settings').hide()
                self.settings.child('camera_settings','image_size','Nx').setValue(self.CCDSIZEX)
                self.settings.child('camera_settings','image_size','Ny').setValue(1)


            elif read_mode == 3: #single track
                self.settings.child('camera_settings', 'readout_settings','mt_settings').hide()
                self.settings.child('camera_settings', 'readout_settings','st_settings').show()
                self.settings.child('camera_settings', 'readout_settings','image_settings').hide()

                err = self.set_single_track_area()

            elif read_mode == 1: #multitrack
                self.settings.child('camera_settings', 'readout_settings','mt_settings').show()
                self.settings.child('camera_settings', 'readout_settings','st_settings').hide()
                self.settings.child('camera_settings', 'readout_settings','image_settings').hide()

                err = self.set_multi_track_area()



            elif read_mode == 2: #random
                err = 'Random mode not implemented yet'
                
            elif read_mode == 4: #image
                self.settings.child('camera_settings', 'readout_settings','mt_settings').hide()
                self.settings.child('camera_settings', 'readout_settings','st_settings').hide()
                self.settings.child('camera_settings', 'readout_settings','image_settings').show()

                self.set_image_area()

                
            elif read_mode == 5: #croped
                err = 'Croped mode not implemented yet'
            self.emit_status(ThreadCommand('Update_Status',[err,'log']))
            
            (err, timings) = self.camera_controller.GetAcquisitionTimings()
            self.settings.child('camera_settings', 'exposure').setValue(timings['exposure']*1000)

            self.x_axis = self.get_xaxis()
            self.y_axis = self.get_yaxis()

    def set_multi_track_area(self):

        N = self.settings.child('camera_settings', 'readout_settings', 'mt_settings', 'mt_N').value()
        height = self.settings.child('camera_settings', 'readout_settings', 'mt_settings', 'mt_height').value()
        offset = self.settings.child('camera_settings', 'readout_settings', 'mt_settings', 'mt_offset').value()
        (err, bottom, gap) = self.camera_controller.SetMultiTrack(N, height, offset)
        self.settings.child('camera_settings', 'readout_settings', 'mt_settings', 'mt_bottom').setValue(bottom)
        self.settings.child('camera_settings', 'readout_settings', 'mt_settings', 'mt_gap').setValue(gap)
        if err == 'DRV_SUCCESS':
            self.settings.child('camera_settings', 'image_size', 'Nx').setValue(self.CCDSIZEX)
            self.settings.child('camera_settings', 'image_size', 'Ny').setValue(N)
        return err

    def set_single_track_area(self):
        center = self.settings.child('camera_settings', 'readout_settings', 'st_settings', 'st_center').value()
        height = self.settings.child('camera_settings', 'readout_settings', 'st_settings', 'st_height').value()
        err = self.camera_controller.SetSingleTrack(center, height)
        if err == 'DRV_SUCCESS':
            self.settings.child('camera_settings', 'image_size', 'Nx').setValue(self.CCDSIZEX)
            self.settings.child('camera_settings', 'image_size', 'Ny').setValue(1)

        return err


    def set_image_area(self):

        binx = self.settings.child('camera_settings', 'readout_settings', 'image_settings', 'bin_x').value()
        biny = self.settings.child('camera_settings', 'readout_settings','image_settings', 'bin_y').value()
        startx = self.settings.child('camera_settings', 'readout_settings', 'image_settings', 'im_startx').value()
        endx = self.settings.child('camera_settings', 'readout_settings', 'image_settings', 'im_endx').value()
        starty = self.settings.child('camera_settings', 'readout_settings', 'image_settings', 'im_starty').value()
        endy = self.settings.child('camera_settings', 'readout_settings', 'image_settings', 'im_endy').value()
        err = self.camera_controller.SetImage(binx, biny, startx, endx, starty, endy)
        if err == 'DRV_SUCCESS':
            self.settings.child('camera_settings', 'image_size', 'Nx').setValue(int((endx-startx+1)/binx))
            self.settings.child('camera_settings', 'image_size', 'Ny').setValue(int((endy-starty+1)/biny))

        return err

    def ini_detector(self, controller=None):
        """
            Initialisation procedure of the detector in four steps :
                * Register callback to get data from camera
                * Get image size and current binning
                * Set and Get temperature from camera
                * Init axes from image

            Returns
            -------
            string list ???
                The initialized status.

            See Also
            --------
            daq_utils.ThreadCommand, hardware1D.DAQ_1DViewer_Picoscope.update_pico_settings
        """

        self.camera_controller = self.ini_detector_init(old_controller=controller,
                                                        new_controller=sdk2.AndorSDK())

        self.emit_status(ThreadCommand('show_splash', "Set/Get Camera's settings"))
        self.ini_camera()

        # %%%%%%% init axes from image
        self.x_axis = self.get_xaxis()
        self.y_axis = self.get_yaxis()

        self.emit_status(ThreadCommand('close_splash'))

        info = ""
        initialized = True
        return info, initialized

    def get_ROI_size_x(self):
        self.CCDSIZEX, self.CCDSIZEY = self.camera_controller.GetDetector()
        return self.CCDSIZEX

    def get_pixel_size(self):
        err, (width, height) = self.camera_controller.GetPixelSize()
        if err == 'DRV_SUCCESS':
            return width, height
        else:
            pass
        return 0., 0.

    def ini_camera(self):
        #  %%%%%% Get image size and current binning
        # get info from camera
        model_param = self.settings.child('camera_settings', 'camera_model')
        cam_index = model_param.opts['limits'].index(model_param.value())
        self.camera_controller.SetCurrentCamera(camera_list[cam_index]['handle'])

        self.CCDSIZEX, self.CCDSIZEY = self.camera_controller.GetDetector()
        self.settings.child('camera_settings', 'readout_settings',
                            'st_settings', 'st_center').setLimits((1, self.CCDSIZEY))
        self.settings.child('camera_settings', 'readout_settings',
                            'st_settings', 'st_height').setLimits((1, self.CCDSIZEY))
        self.settings.child('camera_settings', 'readout_settings', 'image_settings', 'im_endy').setValue(self.CCDSIZEY)
        self.settings.child('camera_settings', 'readout_settings', 'image_settings',
                            'im_endy').setOpts(max=self.CCDSIZEY, default=self.CCDSIZEY)
        self.settings.child('camera_settings', 'readout_settings', 'image_settings', 'im_endx').setValue(self.CCDSIZEX)
        self.settings.child('camera_settings', 'readout_settings', 'image_settings',
                            'im_endx').setOpts(max=self.CCDSIZEX, default=self.CCDSIZEX)

        # get max exposure range
        err, maxexpo = self.camera_controller.GetMaximumExposure()
        if err == 'DRV_SUCCESS':
            self.settings.child('camera_settings', 'exposure').setLimits((0, maxexpo * 1000))

        # set default read mode (full vertical binning)
        self.update_read_mode()

        # %%%%%%% Set and Get temperature from camera
        # get temperature range
        (err, temp_range) = self.camera_controller.GetTemperatureRange()
        if err == "DRV_SUCCESS":
            self.settings.child('camera_settings', 'temperature_settings', 'set_point').setLimits(
                (temp_range[0], temp_range[1]))


        self.set_shutter()

        if not self.camera_controller.IsCoolerOn():  # gets 0 or 1
            self.camera_controller.CoolerON()

        self.camera_controller.SetTemperature(
            self.settings.child('camera_settings', 'temperature_settings', 'set_point').value())
        locked_status, temp = self.camera_controller.GetTemperature()
        self.settings.child('camera_settings', 'temperature_settings', 'current_value').setValue(temp)
        self.settings.child('camera_settings', 'temperature_settings', 'locked').setValue(
            locked_status == 'DRV_TEMP_STABILIZED')
        # set timer to update temperature info from controller
        self.timer.start(2000)

        callback = AndorCallback(self.camera_controller.WaitForAcquisition)
        self.callback_thread = QtCore.QThread()
        callback.moveToThread(self.callback_thread)
        callback.data_sig.connect(self.emit_data)  # when the wait for acquisition returns (with data taken), emit_data will be fired

        self.callback_signal.connect(callback.wait_for_acquisition)
        self.callback_thread.callback = callback
        self.callback_thread.start()


    def set_shutter(self):
        typ = self.settings.child('camera_settings', 'shutter', 'shutter_type').opts['limits'].index(
                                        self.settings.child('camera_settings', 'shutter', 'shutter_type').value())
        mode = self.settings.child('camera_settings', 'shutter', 'shutter_mode').opts['limits'].index(
                                        self.settings.child('camera_settings', 'shutter', 'shutter_mode').value())

        self.camera_controller.SetShutter(typ, mode, self.settings.child('camera_settings', 'shutter', 'shutter_closing_time').value(),
                                          self.settings.child('camera_settings', 'shutter', 'shutter_opening_time').value())

    def updated_timer(self):
        """

        """
        locked_status, temp = self.camera_controller.GetTemperature()
        self.settings.child('camera_settings', 'temperature_settings', 'current_value').setValue(temp)
        self.settings.child('camera_settings', 'temperature_settings', 'locked').setValue(
            locked_status == 'DRV_TEMP_STABILIZED')

    def close(self):
        """

        """
        if self.camera_controller is not None:
            err, temp = self.camera_controller.GetTemperature()
            print(temp)
            if temp < -20.:
                print(
                    "Camera temperature is still at {:d}°C. Closing it now may damage it! The cooling will be maintained "
                    "while shutting down camera. Keep it power plugged!!!".format(
                        temp))
                self.camera_controller.SetCoolerMode(1)
            self.timer.stop()
            self.camera_controller.close()

    def get_xaxis(self):
        """
            Obtain the horizontal axis of the image.

            Returns
            -------
            1D numpy array
                Contains a vector of integer corresponding to the horizontal camera pixels.
        """
        if self.camera_controller is not None:
            # if self.control_type == "camera":
            Nx = self.settings.child('camera_settings', 'image_size', 'Nx').value()
            self.x_axis = Axis(data=np.linspace(0, Nx - 1, Nx, dtype=int), label='Pixels', index=1)

        else:
            raise(Exception('controller not defined'))
        return self.x_axis

    def get_yaxis(self):
        """
            Obtain the vertical axis of the image.

            Returns
            -------
            1D numpy array
                Contains a vector of integer corresponding to the vertical camera pixels.
        """
        if self.camera_controller is not None:

            Ny = self.settings.child('camera_settings', 'image_size', 'Ny').value()
            self.y_axis = Axis(data=np.linspace(0, Ny - 1, Ny, dtype=int), label='Pixels', index=0)
        else:
            raise (Exception('Camera not defined'))
        return self.y_axis

    def prepare_data(self):
        sizex = self.settings.child('camera_settings', 'image_size', 'Nx').value()
        sizey = self.settings.child('camera_settings', 'image_size', 'Ny').value()

        # %%%%%% Initialize data: self.data for the memory to store new data and self.data_average to store the average data
        image_size = sizex * sizey
        self.data = np.zeros((image_size,), dtype=int)
        self.data_pointer = self.data.ctypes.data_as(ctypes.c_void_p)

        data_shape = 'Data2D' if sizey != 1 else 'Data1D'
        if data_shape != self.data_shape:
            self.data_shape = data_shape
            # init the viewers
            self.dte_signal_temp.emit(DataToExport('Camera',
                                                   data=[DataFromPlugins(name='Camera ',
                                                               data=[np.squeeze(
                                                                   self.data.reshape((sizey, sizex)).astype(float))],
                                                               dim=self.data_shape)]))

    def grab_data(self, Naverage=1, **kwargs):
        """
            Start new acquisition in two steps :
                * Initialize data: self.data for the memory to store new data and self.data_average to store the average data
                * Start acquisition with the given exposure in ms, in "1d" or "2d" mode

            =============== =========== =============================
            **Parameters**   **Type**    **Description**
            Naverage         int         Number of images to average
            =============== =========== =============================

            See Also
            --------
            daq_utils.ThreadCommand
        """
        try:
            self.camera_done = False

            self.ind_grabbed = 0  # to keep track of the current image in the average
            self.Naverage = Naverage  #

            self.prepare_data()
            if Naverage == 1:
                self.camera_controller.SetAcquisitionMode(1)
            else:
                self.camera_controller.SetAcquisitionMode(2)
                self.camera_controller.SetNumberAccumulations(Naverage)

            self.camera_controller.SetExposureTime(
                self.settings.child('camera_settings', 'exposure').value() / 1000)  # temp should be in s
            (err, timings) = self.camera_controller.GetAcquisitionTimings()
            self.settings.child('camera_settings', 'exposure').setValue(timings['exposure'] * 1000)
            # %%%%% Start acquisition with the given exposure in ms, in "1d" or "2d" mode
            self.camera_controller.StartAcquisition()
            self.callback_signal.emit(self.Naverage)  # will trigger the waitfor acquisition

        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [str(e), "log"]))

    def stop(self):
        """
            stop the camera's actions.
        """
        try:
            self.camera_controller.CancelWait()  # first cancel the waitacquistion (if any)
            QtWidgets.QApplication.processEvents()
            self.camera_controller.AbortAcquisition()  # abort the camera actions

        except:
            pass
        return ""


class AndorCallback(QtCore.QObject):
    """

    """
    data_sig = QtCore.Signal()

    def __init__(self, wait_fn):
        super(AndorCallback, self).__init__()
        self.wait_fn = wait_fn

    def wait_for_acquisition(self, n_average: int):
        for ind in range(n_average):
            err = self.wait_fn()

        if err != 'DRV_NO_NEW_DATA':  # will be returned if the main thread called CancelWait
            self.data_sig.emit()


if __name__ == '__main__':
    main(__file__, init=True)




#!/usr/bin/python

###############
# andor_sdk3.py
#
# Copyright David Baddeley, 2012
# d.baddeley@auckland.ac.nz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# modified by S. Weber 2020
# PyMoDAQ compatibility
################

from . import sdk3cam
from . import sdk3
import numpy as np
import threading
import ctypes
import time
import traceback
import logging
import queue as Queue


class AndorBase(sdk3cam.SDK3Camera):

    def __init__(self, camNum, nbuffer=1):
        # define properties
        self.CameraAcquiring = sdk3cam.ATBool()
        self.SensorCooling = sdk3cam.ATBool()

        self.AcquisitionStart = sdk3cam.ATCommand()
        self.AcquisitionStop = sdk3cam.ATCommand()

        self.CycleMode = sdk3cam.ATEnum()
        self.ElectronicShutteringMode = sdk3cam.ATEnum()
        self.FanSpeed = sdk3cam.ATEnum()
        self.PreAmpGainChannel = sdk3cam.ATEnum()
        self.PixelEncoding = sdk3cam.ATEnum()
        self.PixelReadoutRate = sdk3cam.ATEnum()
        self.PreAmpGain = sdk3cam.ATEnum()
        self.PreAmpGainSelector = sdk3cam.ATEnum()
        self.TriggerMode = sdk3cam.ATEnum()

        self.FullAOIControl = sdk3cam.ATBool()
        self.AOIHeight = sdk3cam.ATInt()
        self.AOILeft = sdk3cam.ATInt()
        self.AOITop = sdk3cam.ATInt()
        self.AOIWidth = sdk3cam.ATInt()
        self.AOIStride = sdk3cam.ATInt()

        self.FrameCount = sdk3cam.ATInt()
        self.ImageSizeBytes = sdk3cam.ATInt()
        self.SensorHeight = sdk3cam.ATInt()
        self.SensorWidth = sdk3cam.ATInt()

        self.ShutterMode = sdk3cam.ATEnum()
        self.ShutterOutputMode = sdk3cam.ATEnum()

        self.CameraModel = sdk3cam.ATString()
        self.CameraName = sdk3cam.ATString()
        self.ControllerID = sdk3cam.ATString()
        self.SerialNumber = sdk3cam.ATString()

        self.ExposureTime = sdk3cam.ATFloat()
        self.FrameRate = sdk3cam.ATFloat()
        self.SensorTemperature = sdk3cam.ATFloat()
        self.TargetSensorTemperature = sdk3cam.ATFloat()

        sdk3cam.SDK3Camera.__init__(self, camNum)


    def init_camera(self):
        super().init_camera()

    def _flush(self):
        sdk3.Flush(self.handle)

    def queue_single_buffer(self, buf):
        sdk3.QueueBuffer(self.handle, buf.ctypes.data_as(sdk3.POINTER(sdk3.AT_U8)), buf.nbytes)


    def wait_buffer(self):
        try:
            pData, lData = sdk3.WaitBuffer(self.handle, 0)
        except sdk3.TimeoutError as e:
            return
        except sdk3.CameraError as e:
            #    if not e.errNo == sdk3.AT_ERR_NODATA:
            #        traceback.print_exc()
            return
        return ctypes.addressof(pData.contents)

    def get_image_fom_buffer(self, Ny, Nx, buffer):
        data = np.empty((Ny, Nx), np.uint16)

        a_s = self.AOIStride.getValue()
        dt = self.PixelEncoding.getString()
        sdk3.ConvertBuffer(buffer.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),
                           data.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),
                           Ny, Nx, a_s, dt, 'Mono16')

        return data


    def flush(self):
        self._flush()

    def GetSerialNumber(self):
        return self.SerialNumber.getValue()

    def SetIntegTime(self, iTime):
        """Set Integration time in milliseconds"""
        self.ExposureTime.setValue(iTime * 1e-3)
        self.FrameRate.setValue(self.FrameRate.max())

    def GetIntegTime(self):
        """Return Integration Time in milliseconds"""
        return self.ExposureTime.getValue() * 1000

    def GetCCDWidth(self):
        return self.SensorWidth.getValue()

    def GetCCDHeight(self):
        return self.SensorHeight.getValue()

    def GetCCDTemp(self):
        # for some reason querying the temperature takes a lot of time - do it less often
        # return self.SensorTemperature.getValue()

        return self._temp

    def GetPicWidth(self):
        return self.AOIWidth.getValue()

    def GetPicHeight(self):

        return self.AOIHeight.getValue()

    def SetROIIndex(self, index):
        width, height, top, left = self.validROIS[index]

        self.AOIWidth.setValue(width)
        self.AOILeft.setValue(left)
        self.AOIHeight.setValue(height)
        self.AOITop.setValue(top)

    def StopAq(self):
        if self.CameraAcquiring.getValue():
            self.AcquisitionStop()

    def GenStartMetadata(self, mdh):
        if self.active:
            self.GetStatus()

            mdh.setEntry('Camera.Name', 'Andor Neo')

            mdh.setEntry('Camera.IntegrationTime', self.GetIntegTime())
            mdh.setEntry('Camera.CycleTime', self.GetIntegTime())
            mdh.setEntry('Camera.EMGain', 1)

            mdh.setEntry('Camera.ROIPosX', self.GetROIX1())
            mdh.setEntry('Camera.ROIPosY',  self.GetROIY1())
            mdh.setEntry('Camera.ROIWidth', self.GetROIX2() - self.GetROIX1())
            mdh.setEntry('Camera.ROIHeight',  self.GetROIY2() - self.GetROIY1())
            #mdh.setEntry('Camera.StartCCDTemp',  self.GetCCDTemp())

            mdh.setEntry('Camera.ReadNoise', 1)
            mdh.setEntry('Camera.NoiseFactor', 1)
            mdh.setEntry('Camera.ElectronsPerCount', 1)
            #mdh.setEntry('Camera.ADOffset', self.noiseMaker.ADOffset)

            #mdh.setEntry('Simulation.Fluorophores', self.fluors.fl)
            #mdh.setEntry('Simulation.LaserPowers', self.laserPowers)

            #realEMGain = ccdCalibrator.getCalibratedCCDGain(self.GetEMGain(), self.GetCCDTempSetPoint())
            # if not realEMGain == None:
            mdh.setEntry('Camera.TrueEMGain', 1)


    def GetFPS(self):
        return self.FrameRate.getValue()


def getCameraInfos():
    """
    Gives list of connected camera models, names and serial_numbers (including simulation ones)
    Returns
    -------
    models: (list) list of string
    names: (list) list of string
    serial_numbers: (list) list of string
    """
    models = []
    names = []
    serial_numbers = []
    for ind in range(sdk3cam.getNumCameras()):
        cam = AndorBase(ind)
        sdk3cam.camReg.regCamera()
        cam.init_camera()
        try:
            models.append(cam.CameraModel.getValue())
        except:
            models.append('')
        try:
            names.append(cam.CameraName.getValue())
        except:
            names.append('')
        try:
            serial_numbers.append(cam.SerialNumber.getValue())
        except:
            serial_numbers.append('')
        cam.shutdown()
        sdk3cam.camReg.unregCamera()
    return models, names, serial_numbers


class AndorCamera(AndorBase):
    def __init__(self, camNum, nbuffer=1):
        # define properties
        self.Overlap = sdk3cam.ATBool()
        self.SpuriousNoiseFilter = sdk3cam.ATBool()

        self.CameraDump = sdk3cam.ATCommand()
        self.SoftwareTrigger = sdk3cam.ATCommand()

        self.TemperatureControl = sdk3cam.ATEnum()
        self.TemperatureStatus = sdk3cam.ATEnum()
        self.SimplePreAmpGainControl = sdk3cam.ATEnum()

        self.BitDepth = sdk3cam.ATEnum()

        self.ActualExposureTime = sdk3cam.ATFloat()
        self.BurstRate = sdk3cam.ATFloat()
        self.ReadoutTime = sdk3cam.ATFloat()
        self.ExternalTriggerDelay = sdk3cam.ATFloat()

        self.AOIBinning = sdk3cam.ATEnum()
        self.AOIHBin = sdk3cam.ATInt()
        self.AOIVBin = sdk3cam.ATInt()

        self.PixelWidth = sdk3cam.ATFloat()
        self.PixelHeight = sdk3cam.ATFloat()

        self.AccumulateCount = sdk3cam.ATInt()
        self.BaselineLevel = sdk3cam.ATInt()
        self.BurstCount = sdk3cam.ATInt()
        self.LUTIndex = sdk3cam.ATInt()
        self.LUTValue = sdk3cam.ATInt()

        self.ControllerID = sdk3cam.ATString()
        self.FirmwareVersion = sdk3cam.ATString()

        super().__init__(camNum, nbuffer)


class AndorSim(AndorBase):
    def __init__(self, camNum):
        # define properties
        self.SynchronousTriggering = sdk3cam.ATBool()

        self.PixelCorrection = sdk3cam.ATEnum()
        self.TriggerSelector = sdk3cam.ATEnum()
        self.TriggerSource = sdk3cam.ATEnum()

        self.PixelHeight = sdk3cam.ATFloat()
        self.PixelWidth = sdk3cam.ATFloat()

        self.AOIHBin = sdk3cam.ATInt()
        self.AOIVbin = sdk3cam.ATInt()

        super().__init__(camNum)



import numpy as np
from qtpy import QtWidgets
from time import perf_counter

from pymodaq_utils.utils import ThreadCommand, find_dict_in_list_from_key_val
from pymodaq_utils.logger import set_logger, get_module_name

from pymodaq_gui.parameter import utils as putils

from pymodaq.utils.data import Axis, DataFromPlugins, DataToExport
from pymodaq.control_modules.viewer_utility_classes import main, DAQ_Viewer_base, comon_parameters


from pymodaq_plugins_andor.daq_viewer_plugins.plugins_2D.daq_2Dviewer_AndorSCMOS import DAQ_2DViewer_AndorSCMOS
from pymodaq_plugins_andor.daq_move_plugins.daq_move_Shamrock import DAQ_Move_Shamrock

logger = set_logger(get_module_name(__file__))


class DAQ_1DViewer_ShamrockCCDComposition(DAQ_2DViewer_AndorSCMOS):
    """
        =============== ==================

        =============== ==================

        See Also
        --------
        utility_classes.DAQ_Viewer_base
    """
    param_camera = DAQ_2DViewer_AndorSCMOS.params
    params_shamrock = DAQ_Move_Shamrock.params
    putils.get_param_dict_from_name(params_shamrock, 'andor_lib', pop=True)

    d = putils.get_param_dict_from_name(params_shamrock, 'spectro_wl')
    if d is not None:
        d['readonly'] = False
    d = putils.get_param_dict_from_name(params_shamrock, 'flip_wavelength')
    if d is not None:
        d['visible'] = True

    params = [{'title': 'Get Calibration:', 'name': 'get_calib', 'type': 'bool_push', 'value': False,
              'label': 'Update!'},] + param_camera + [
              {'title': 'Shamrock Settings:', 'name': 'sham_settings', 'type': 'group', 'children': params_shamrock},
    ]


    def ini_attributes(self):
        self.controller: DAQ_2DViewer_AndorSCMOS = None
        self.shamrock_controller: DAQ_Move_Shamrock = None

        self.x_axis: Axis = None
        self.is_calibrated = False

        super().ini_attributes()

    def commit_settings(self, param):

        if param.name() == 'flip_wavelength':
            self.get_xaxis()
        elif 'camera_settings' in putils.get_param_path(param):
            super().commit_settings(param)
        elif 'spectro_settings' in putils.get_param_path(param):
            self.shamrock_controller.commit_settings(param)
        QtWidgets.QApplication.processEvents()
        if param.name() == 'spectro_wl':
            self.is_calibrated = False
            self.get_xaxis()
        elif param.name() == 'zero_order':
            self.is_calibrated = False
        elif param.name() == 'flip_wavelength':
            self.get_xaxis()
        elif param.name() == 'readout' or param.name() in \
            putils.iter_children(self.settings.child('camera_settings', 'readout_settings')):
            self.get_xaxis()
        elif param.name() == 'get_calib':
            if param.value():
                self.get_xaxis()
                param.setValue(False)

    def ini_detector(self, controller=None):
        cam_status, cam_init = super().ini_detector(controller)
        QtWidgets.QApplication.processEvents()

        self.shamrock_controller = DAQ_Move_Shamrock(None, self.settings.child('sham_settings').saveState())
        self.shamrock_controller.settings = self.settings.child('sham_settings')
        self.shamrock_controller.emit_status = self.emit_status
        sham_status, sham_init = self.shamrock_controller.ini_stage(controller)

        QtWidgets.QApplication.processEvents()


        initialized = sham_init and cam_init

        self.setCalibration()
        return sham_status + cam_status, initialized

    def setCalibration(self):
        #setNpixels
        width, height = self.get_pixel_size()
        err = self.shamrock_controller.controller.SetNumberPixelsSR(0, self.get_ROI_size_x())
        err = self.shamrock_controller.controller.SetPixelWidthSR(0, width)

        self.shamrock_controller.get_wavelength()
        self.x_axis = self.get_xaxis()


    def getCalibration(self):

        if self.shamrock_controller is not None and self.shamrock_controller.controller is not None:
            (err, calib) = self.shamrock_controller.controller.GetCalibrationSR(0, self.get_ROI_size_x())
            if err != "SHAMROCK_SUCCESS":
                raise Exception(err)

            calib = np.array(calib)
        else:
            raise(Exception('controller not defined'))
        self.is_calibrated = True
        return calib

    def get_xaxis(self):
        """
            Obtain the horizontal axis of the image.

            Returns
            -------
            1D numpy array
                Contains a vector of integer corresponding to the horizontal camera pixels.
        """

        if self.shamrock_controller is None or np.abs(self.settings.child('sham_settings', 'spectro_settings', 'spectro_wl').value()) < 1e-3:
            nx = self.get_ROI_size_x()
            calib = np.linspace(0, nx, nx-1)
            self.x_axis = Axis(data=calib, label='Wavelength (nm)')
        else:
            calib = self.getCalibration()

            if (calib.astype('int') != 0).all():  # check if calib values are equal to zero
                if self.settings.child('sham_settings', 'spectro_settings', 'flip_wavelength').value():
                    calib = calib[::-1]

            else:
                self.settings.child('sham_settings', 'spectro_settings', 'flip_wavelength').setValue(False)
                #self.emit_status(ThreadCommand('Update_Status', ['Impossible to flip wavelength', "log"]))

            self.x_axis = Axis(data=calib, label='Wavelength (nm)')
        return self.x_axis

    def get_exposure_ms(self):
        #for compatibility with PyMoDAQ Spectro module
        self.emit_status(ThreadCommand('exposure_ms', [self.settings.child('camera_settings', 'exposure').value()]))

    def set_exposure_ms(self, exposure):
        self.settings.child('camera_settings', 'exposure').setValue(exposure)
        QtWidgets.QApplication.processEvents()
        self.emit_status(ThreadCommand('exposure_ms', [self.settings.child('camera_settings', 'exposure').value()]))

    def stop(self):
        if self.controller is not None:
            super().stop()
        if self.shamrock_controller is not None:
            self.shamrock_controller.stop()

    def close(self):
        self.stop()
        if self.shamrock_controller is not None:
            self.shamrock_controller.close()
        super().close()

    def grab_data(self, Naverage=1, **kwargs):
        if not self.is_calibrated:
            self.get_xaxis()
        super().grab_data( Naverage, **kwargs)

    def emit_data(self, buffer_pointer):
        """
            overloadded function from DAQ_2DViewer_AndorScmos
        """
        try:
            buff_temp = buffer_pointer[0]
            self.current_buffer += 1
            self.n_grabed_data += 1
            self.n_grabed_frame_rate += 1
            #print(f'ind_grabemit:{self.n_grabed_data}')
            self.current_buffer = self.current_buffer % self._Nbuffers
            #print(f'ind_current_buffer:{self.current_buffer}')

            if self.buffers[self.current_buffer].ctypes.data != buff_temp:
                # buff_val = [self.buffers[ind].ctypes.data for ind in range(len(self.buffers))].index(buff_temp)
                # print(f'Buffer index should be {self.current_buffer} but is in fact {buff_val}')
                self.stop()
                QtWidgets.QApplication.processEvents()

                self.emit_status(ThreadCommand('Update_Status',
                                               ['Returned buffer not equal to expected buffer,'
                                                ' restarting acquisition and'
                                                ' freeing buffers', 'log']))
                logger.warning('Returned buffer not equal to expected buffer, restarting acquisition and'
                               ' freeing buffers')
                self._reset_buffers_cmd = True
                self.grab_data(self.Naverage, live=self.live, wait_time=self.wait_time)
                return

            cam_name = self.settings.child('camera_settings', 'camera_model').value()
            Nx = self.settings.child('camera_settings', 'image_settings', 'im_width').value()
            Ny = self.settings.child('camera_settings', 'image_settings', 'im_height').value()
            data = self.camera_controller.get_image_fom_buffer(Nx, Ny, self.buffers[self.current_buffer]).T

            if self.n_grabed_data % self.Naverage == 0 and self.live:
                self.data = 1 / self.Naverage * data
            else:
                self.data += 1 / self.Naverage * data

            if not self.live:
                if self.n_grabed_data > self.Naverage:
                    self.stop()
                else:
                    #self.data += 1 / self.Naverage * data
                    if self.n_grabed_data == self.Naverage:
                        self.data_grabed_signal.emit([
                            DataFromPlugins(name=cam_name, data=[self.data], dim=self.data_shape,
                                            x_axis=self.x_axis)])

                    elif self.n_grabed_data < self.Naverage:
                        self.data_grabed_signal_temp.emit([
                            DataFromPlugins(name=cam_name, data=[self.data * self.Naverage / self.n_grabed_data],
                                            dim=self.data_shape,
                                            x_axis=self.x_axis)])
            else:  # in live mode
                if perf_counter() - self.start_time > self.refresh_time_fr / 1000:  # refresh the frame rate every
                    # refresh_time_fr ms
                    self.settings.child('camera_settings',
                                        'frame_rate').setValue(self.n_grabed_frame_rate / (self.refresh_time_fr / 1000))
                    self.start_time = perf_counter()
                    self.n_grabed_frame_rate = 0

                if self.n_grabed_data % self.Naverage == 0:
                    self.data_grabed_signal.emit([
                        DataFromPlugins(name=cam_name, data=[self.data], dim=self.data_shape,
                                        x_axis=self.x_axis)])
                else:
                    if self.n_grabed_data % self.Naverage != 0:
                        n_grabed = self.n_grabed_data % self.Naverage
                    else:
                        n_grabed = self.Naverage
                    self.data_grabed_signal_temp.emit([
                        DataFromPlugins(name=cam_name,
                                        data=[self.data * self.Naverage / n_grabed],
                                        dim=self.data_shape,
                                        x_axis=self.x_axis)])

            self.camera_controller.queue_single_buffer(self.buffers[self.current_buffer])

        except Exception as e:
            logger.exception(str(e))



if __name__ == '__main__':
    main(__file__, True)




from easydict import EasyDict as edict
from pymodaq.control_modules.move_utility_classes import DAQ_Move_base, comon_parameters_fun, main
from pymodaq_utils.utils import ThreadCommand
from pymodaq_plugins_andor.hardware import shamrock_sdk


libpath = shamrock_sdk.dllpath


class DAQ_Move_Shamrock(DAQ_Move_base):

    _controller_units = 'nm'
    is_multiaxes = False
    axes_names = []  # "list of strings of the multiaxes
    _epsilon = 0.1

    params = [
        {'title': 'Dll library:', 'name': 'andor_lib', 'type': 'browsepath', 'value': str(libpath), 'readonly': True},
        {'title': 'Spectro Settings:', 'name': 'spectro_settings', 'type': 'group', 'expanded': True,
            'children': [
                {'title': 'Spectro SN:', 'name': 'spectro_serialnumber', 'type': 'str', 'value': '',
                    'readonly': True},
                {'title': 'Wavelength (nm):', 'name': 'spectro_wl', 'type': 'float', 'value': 600, 'min': 0,
                    'readonly': True},
                {'title': 'Home Wavelength (nm):', 'name': 'spectro_wl_home', 'type': 'float', 'value': 600, 'min': 0,
                 'readonly': False},
                {'title': 'Grating Settings:', 'name': 'grating_settings', 'type': 'group', 'expanded': True,
                    'children': [
                        {'title': 'Grating:', 'name': 'grating', 'type': 'list'},
                        {'title': 'Lines (/mm):', 'name': 'lines', 'type': 'int', 'readonly': True},
                        {'title': 'Blaze WL (nm):', 'name': 'blaze', 'type': 'str', 'readonly': True},
                    ]},
                {'title': 'Flip wavelength axis:', 'name': 'flip_wavelength', 'type': 'bool', 'value': False,
                    'visible': False},
                {'title': 'Go to zero order:', 'name': 'zero_order', 'type': 'bool'},
            ]},
        ] + comon_parameters_fun(is_multiaxes, axes_names, epsilon=_epsilon)

    def commit_settings(self, param):
        """
            | Activate parameters changes on the hardware from parameter's name.
            |

            =============== ================================    =========================
            **Parameters**   **Type**                           **Description**
            *param*          instance of pyqtgraph parameter    The parameter to activate
            =============== ================================    =========================

            Three profile of parameter :
                * **bin_x** : set binning camera from bin_x parameter's value
                * **bin_y** : set binning camera from bin_y parameter's value
                * **set_point** : Set the camera's temperature from parameter's value.

        """
        try:
            if param.name() == 'grating':
                index_grating = self.grating_list.index(param.value())
                self.get_set_grating(index_grating)
                self.set_wavelength(self.settings.child('spectro_settings', 'spectro_wl').value())

            elif param.name() == 'spectro_wl':
                self.set_wavelength(param.value())

            elif param.name() == 'zero_order':
                if param.value():
                    param.setValue(False)
                    self.emit_status(ThreadCommand('show_splash', "Moving to zero order, please wait!"))
                    err = self.shamrock_controller.GotoZeroOrderSR(0)
                    if err != 'SHAMROCK_SUCCESS':
                        raise Exception(err)
                    self.check_position()
                    self.emit_status(ThreadCommand('close_splash'))

        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [str(e), 'log']))

    def ini_stage(self, controller=None):
        """Actuator communication initialization

        Parameters
        ----------
        controller: (object) custom object of a PyMoDAQ plugin (Slave case). None if only one actuator by controller (Master case)

        Returns
        -------
        self.status (edict): with initialization status: three fields:
            * info (str)
            * controller (object) initialized controller
            *initialized: (bool): False if initialization failed otherwise True
        """
        self.shamrock_controller = self.ini_stage_init(old_controller=controller,
                                                       new_controller=shamrock_sdk.ShamrockSDK())

        self.emit_status(ThreadCommand('show_splash', "Set/Get Shamrock's settings"))
        self.ini_spectro()

        initialized = True
        self.emit_status(ThreadCommand('close_splash'))
        return '', initialized

    def get_actuator_value(self):
        """Get the current position from the hardware with scaling conversion.

        Returns
        -------
        float: The position obtained after scaling conversion.
        """
        pos = self.get_wavelength()
        ##

        pos = self.get_position_with_scaling(pos)
        return pos

    def move_abs(self, position):
        """ Move the actuator to the absolute target defined by position

        Parameters
        ----------
        position: (flaot) value of the absolute target positioning
        """

        position = self.check_bound(position)  # if user checked bounds, the defined bounds are applied here
        position = self.set_position_with_scaling(position)  # apply scaling if the user specified one

        self.set_wavelength(position)

        ##############################

        self.target_position = position
        self.poll_moving()  # start a loop to poll the current actuator value and compare it with target position

    def move_rel(self, position):
        """ Move the actuator to the relative target actuator value defined by position

        Parameters
        ----------
        position: (flaot) value of the relative target positioning
        """
        position = self.check_bound(self.current_position + position) - self.current_position
        self.target_position = position + self.current_position

        self.set_wavelength(self.target_position)
        ##############################

        self.poll_moving()

    def move_home(self):
        """

        """
        self.move_abs(self.settings.child('spectro_settings', 'spectro_wl_home').value())

    def stop_motion(self):
        """
        Call the specific move_done function (depending on the hardware).

        See Also
        --------
        move_done
        """

        self.move_done()  # to let the interface know the actuator stopped. Direct call as the setwavelength call is
        # blocking anyway

    def close(self):
        """

        """
        if self.shamrock_controller is not None:
            self.shamrock_controller.close()

    def set_wavelength(self, wavelength):
        self.emit_status(ThreadCommand('show_splash', "Setting wavelength, please wait!"))
        err = self.shamrock_controller.SetWavelengthSR(0, wavelength)
        self.emit_status(ThreadCommand('close_splash'))

        if err != 'SHAMROCK_SUCCESS':
            raise IOError(err)

        self.get_wavelength()

    def get_wavelength(self):
        err, wl = self.shamrock_controller.GetWavelengthSR(0)
        if err == "SHAMROCK_SUCCESS":
            self.settings.child('spectro_settings', 'spectro_wl').setValue(wl)
        return float(wl)

    def ini_spectro(self):
        self.settings.child('spectro_settings', 'spectro_serialnumber').setValue(
            self.shamrock_controller.GetSerialNumberSR(0)[1].decode())

        # get grating info
        (err, Ngratings) = self.shamrock_controller.GetNumberGratingsSR(0)
        self.grating_list = []
        for ind_grating in range(1, Ngratings + 1):
            (err, lines, blaze, home, offset) = self.shamrock_controller.GetGratingInfoSR(0, ind_grating)
            self.grating_list.append(str(int(lines)))

        self.settings.child('spectro_settings', 'grating_settings', 'grating').setLimits(self.grating_list)
        err, ind_grating = self.shamrock_controller.GetGratingSR(0)
        self.settings.child('spectro_settings', 'grating_settings', 'grating').setValue(
            self.grating_list[ind_grating - 1])

        self.get_set_grating(ind_grating - 1)

    def get_set_grating(self, ind_grating):
        """
        set the current grating to ind_grating+1. ind_grating corresponds to the index in the GUI graitng list while the SDK index starts at 1...

        """
        self.emit_status(ThreadCommand('show_splash', "Moving grating please wait"))
        err = self.shamrock_controller.SetGratingSR(0, ind_grating + 1)
        err, ind_grating = self.shamrock_controller.GetGratingSR(0)

        (err, lines, blaze, home, offset) = self.shamrock_controller.GetGratingInfoSR(0, ind_grating)
        self.settings.child('spectro_settings', 'grating_settings', 'grating').setValue(
            self.grating_list[ind_grating - 1])
        self.settings.child('spectro_settings', 'grating_settings', 'lines').setValue(lines)
        self.settings.child('spectro_settings', 'grating_settings', 'blaze').setValue(blaze)

        (err, wl_min, wl_max) = self.shamrock_controller.GetWavelengthLimitsSR(0, ind_grating)

        if err == "SHAMROCK_SUCCESS":
            self.settings.child('spectro_settings',
                                'spectro_wl').setOpts(limits=(wl_min, wl_max),
                                                      tip=f'Possible values are within {wl_min} and {wl_max} for'
                                                          f' the selected grating')
            self.settings.child('spectro_settings',
                                'spectro_wl_home').setOpts(limits=(wl_min, wl_max),
                                                           tip=f'Possible values are within {wl_min} and {wl_max} for'
                                                               f' the selected grating')

        self.emit_status(ThreadCommand('close_splash'))

    def stop(self):
        pass


if __name__ == '__main__':
    main(__file__, True)
